#!/usr/bin/env node
/* eslint-disable no-console */
const { omit } = require('lodash');
const Table = require('cli-table3');
const filesize = require('filesize');
const yargs = require('yargs');
const ora = require('ora');
const run = require('..');

require('dotenv').config();

const { argv } = yargs
  .env('ASSETS')
  .usage('$0 <url>', 'Run assets metrics', (y) => y.positional('url', {
      describe: 'Website URL',
      type: 'string',
    }),)
  .array('mimeTypes')
  .option('mimeTypes', {
    alias: 't',
    describe: 'Filter assets based on mime type',
    default: ['javascript'],
  })
  .option('internalPattern', {
    alias: 'i',
    describe: 'Sets assets as internal scripts based on regex pattern',
  });

const { url, ...args } = argv;
const options = omit(args, ['help', 'version', '$0', '_']);
const loading = ora('Getting metrics...').start();

run(url, options)
  .then(
    ({
      assets,
      encodedSize,
      size,
      count,
      internalSize,
      externalSize,
      internalEncodedSize,
      externalEncodedSize,
      internalCount,
      externalCount,
    }) => {
      loading.stop();

      const assetsTable = new Table({
        head: ['URL', 'Encoded Size', 'Size', 'Mime Type', 'Type'],
        wordWrap: true,
      });

      Object.keys(assets).forEach((asset) => {
        const metrics = assets[asset];
        assetsTable.push([
          asset,
          filesize(metrics.encodedSize),
          filesize(metrics.size),
          metrics.mimeType,
          metrics.type,
        ]);
      });
      console.log(String(assetsTable));

      const summaryTable = new Table({
        head: ['Summary', 'Internal', 'External', 'Total'],
      });
      summaryTable.push(['Count', internalCount, externalCount, count]);
      summaryTable.push([
        'Size',
        filesize(internalSize),
        filesize(externalSize),
        filesize(size),
      ]);
      summaryTable.push([
        'Encoded Size',
        filesize(internalEncodedSize),
        filesize(externalEncodedSize),
        filesize(encodedSize),
      ]);
      console.log(String(summaryTable));
    },
  )
  .catch((err) => {
    loading.stop();
    console.error('Error occurred', err);
  });
