const puppeteer = require('puppeteer');
const { get } = require('lodash');

function sanitizeUrl(url) {
  if (!/^(?:f|ht)tps?:\/\//.test(url)) {
    return `https://${url}`;
  }
  return url;
}

const IGNORE_ASSET_REGEX = /(^data:)/;

async function run(plainUrl, options = {}) {
  const pageUrl = sanitizeUrl(plainUrl);
  const { mimeTypes = ['javascript'], internalPattern } = options;

  const browser = await puppeteer.launch({
    headless: true,
    args: ['--no-sandbox', '--disable-setuid-sandbox'],
  });
  const page = await browser.newPage();

  const session = await page.target().createCDPSession();
  await session.send('Network.enable');

  const responses = {};
  session.on('Network.responseReceived', (event) => {
    responses[event.requestId] = event.response;
  });

  const assets = {};
  session.on('Network.dataReceived', (event) => {
    const { url, mimeType } = responses[event.requestId];

    if (
      IGNORE_ASSET_REGEX.test(url)
      || mimeTypes.some((type) => !new RegExp(type).test(mimeType))
    ) {
      return;
    }

    const isInternal =      url.startsWith(pageUrl)
      || (internalPattern ? new RegExp(internalPattern).test(url) : false);
    const asset = assets[url];
    assets[url] = {
      mimeType,
      type: isInternal ? 'internal' : 'external',
      encodedLength: get(asset, 'encodedLength', 0) + event.encodedDataLength,
      length: get(asset, 'length', 0) + event.dataLength,
    };
  });

  await page.goto(pageUrl, {
    waitUntil: 'networkidle0',
  });

  await browser.close();

  let count = 0;
  let internalCount = 0;
  let externalCount = 0;
  let totalLength = 0;
  let totalEncodedLength = 0;
  Object.values(assets).forEach((asset) => {
    count += 1;
    totalLength += asset.length;
    totalEncodedLength += asset.encodedLength;
    if (asset.type === 'internal') {
      internalCount += 1;
    } else {
      externalCount += 1;
    }
  });

  return {
    assets,
    totalLength,
    totalEncodedLength,
    internalCount,
    externalCount,
    count,
  };
}

module.exports = run;
