/* eslint-disable no-console */
const express = require('express');
const bodyParser = require('body-parser');
const Cache = require('ttl-cache');
const promClient = require('prom-client');
const yamlJs = require('yamljs');
const {
 get, merge, fromPairs, isArray 
} = require('lodash');
const extractMetrics = require('..');

function getEnv(key, fallback) {
  return get(process.env, key, fallback);
}

let config = {
  path: getEnv('METRICS_PATH', '/metrics'),
  cache: {
    enabled: getEnv('CACHE_ENABLED', true),
    ttl: getEnv('CACHE_TTL', 300),
    interval: getEnv('CACHE_INTERVAL', 60),
  },
  metricName: getEnv('METRIC_NAME', 'puppeteer_assets'),
  enableOnDemandQuery: getEnv('METRIC_ON_DEMAND_QUERY', false),
  defaults: {
    metrics: {
      file: getEnv('METRIC_FILE_ENABLED', true),
      count: getEnv('METRIC_COUNT_ENABLED', true),
      size: getEnv('METRIC_SIZE_ENABLED', true),
      gzip: getEnv('METRIC_GZIP_ENABLED', true),
      countByMimeType: getEnv('METRIC_COUNT_BY_MIME_TYPE_ENABLED', true),
      sizeByMimeType: getEnv('METRIC_SIZE_BY_MIME_TYPE_ENABLED', true),
      gzipSizeByMimeType: getEnv('METRIC_GZIP_SIZE_BY_MIME_TYPE_ENABLED', true),
    },
    mimeTypePatterns: [getEnv('METRIC_MIME_TYPE_PATTERN', 'javascript')],
    internalPatterns: [getEnv('METRIC_INTERNAL_PATTERN', '^(/|./)')],
  },
  configurations: [],
};

try {
  config = merge(
    {},
    config,
    yamlJs.load(process.env.CONFIG_PATH || `${__dirname}/../config.yml`) || {},
  );
  console.info('> Using configuration file...');
} catch (e) {
  config.enableOnDemandQuery = true;
  console.info('> No configuration file. On demand query enabled.');
}

function getConf(key, conf = {}) {
  return get(conf, key, get(config.defaults, key));
}

const app = express();
app.use(bodyParser.json());

const cache = config.cache.enabled
  ? new Cache({
    ttl: config.cache.ttl,
    interval: config.cache.interval,
  })
  : null;

let running = false;

function parseLabelsParam(arr) {
  return isArray(arr) && arr.length
    ? fromPairs(arr.map((l) => l.split(':'))) || {}
    : {};
}

function iterateToGauge(metrics, field, gauge, page) {
  Object.entries(metrics[field]).forEach(([type, data]) => {
    if (type === 'total') return;
    Object.entries(data).forEach(([mimeType, value]) => {
      if (mimeType === 'total') {
        gauge.set({ page, type }, value);
      }
    });
  });
}

function iterateToGaugeByMimeType(metrics, field, gauge, page) {
  Object.entries(metrics[field]).forEach(([type, data]) => {
    if (type === 'total') return;
    Object.entries(data).forEach(([mimeType, value]) => {
      if (mimeType === 'total') return;
      gauge.set({ page, type, mime_type: mimeType }, value);
    });
  });
}

app.get(
  config.path,
  (() => {
    const { metricName } = config;

    new promClient.Gauge({
      name: 'up',
      help: '1 = up, 0 = not up',
    }).set(1);

    const gaugeFile = getConf('metrics.file')
      ? new promClient.Gauge({
        name: `${metricName}_file`,
        help: 'historically found assets on scrap',
        labelNames: ['page', 'type', 'mime_type', 'url'],
      })
      : null;

    const gaugeCount = getConf('metrics.count')
      ? new promClient.Gauge({
        name: `${metricName}_count`,
        help: 'count of internal assets',
        labelNames: ['page', 'type'],
      })
      : null;

    const gaugeSize = getConf('metrics.size')
      ? new promClient.Gauge({
        name: `${metricName}_size`,
        help: 'real size of assets',
        labelNames: ['page', 'type'],
      })
      : null;

    const gaugeGzip = getConf('metrics.gzip')
      ? new promClient.Gauge({
        name: `${metricName}_gzip`,
        help: 'encoded (gzip) size of assets',
        labelNames: ['page', 'type'],
      })
      : null;

    const gaugeCountByMimeType = getConf('metrics.countByMimeType')
      ? new promClient.Gauge({
        name: `${metricName}_count_by_mime_type`,
        help: 'count of internal assets by mime type',
        labelNames: ['page', 'type', 'mime_type'],
      })
      : null;

    const gaugeSizeByMimeType = getConf('metrics.sizeByMimeType')
      ? new promClient.Gauge({
        name: `${metricName}_size_by_mime_type`,
        help: 'real size of assets by mime type',
        labelNames: ['page', 'type', 'mime_type'],
      })
      : null;

    const gaugeGzipByMimeType = getConf('metrics.gzipByMimeType')
      ? new promClient.Gauge({
        name: `${metricName}_gzip_by_mime_type`,
        help: 'encoded (gzip) size of assets by mime type',
        labelNames: ['page', 'type', 'mime_type'],
      })
      : null;

    const scrapPage = async (route) => {
      const { url: page, ...options } = route;
      let metrics = config.cache.enabled ? cache.get(page) : null;

      if (!metrics) {
        if (!running) {
          running = true;

          promClient.register.setDefaultLabels(options.labels);

          try {
            metrics = await extractMetrics(page, options);
          } catch (e) {
            console.error(e);
          }

          running = false;
          if (config.cache.enabled) cache.set(page, metrics);
        } else {
          return;
        }
      }

      if (getConf('metrics.file')) {
        // file metrics
        const ts = Date.now();
        Object.entries(metrics.assets).forEach(([url, data]) => {
          const { type, mimeType } = data;
          gaugeFile.set(
            {
              page,
              url,
              type,
              mime_type: mimeType,
            },
            ts,
          );
        });
      }

      // default
      if (getConf('metrics.count')) {
        iterateToGauge(metrics, 'count', gaugeCount, page);
      }
      if (getConf('metrics.size')) {
        iterateToGauge(metrics, 'size', gaugeSize, page);
      }
      if (getConf('metrics.gzip')) {
        iterateToGauge(metrics, 'gzip', gaugeGzip, page);
      }

      // by mime type
      if (getConf('metrics.countByMimeType')) {
        iterateToGaugeByMimeType(metrics, 'count', gaugeCountByMimeType, page);
      }
      if (getConf('metrics.sizeByMimeType')) {
        iterateToGaugeByMimeType(metrics, 'size', gaugeSizeByMimeType, page);
      }
      if (getConf('metrics.gzipByMimeType')) {
        iterateToGaugeByMimeType(metrics, 'gzip', gaugeGzipByMimeType, page);
      }
    };

    async function middleware(req, res, next) {
      if (req.query.url) {
        if (!config.enableOnDemandQuery) {
          res.status(400).send('On Demand is disabled');
          return;
        }
      } else if (config.enableOnDemandQuery && !config.configurations.length) {
        res.status(400).send('No URL query argument');
        return;
      } else if (!config.configurations.length) {
        res.status(400).send('No configurations');
        return;
      }

      res.writeHead(200, { 'Content-Type': 'text/plain' });

      if (req.query.url) {
        await scrapPage(
          merge({}, config, {
            url: req.query.url,
            labels: parseLabelsParam(req.query.labels || []),
          }),
        );
      } else {
        await Promise.all(
          config.configurations.map(async (route) => scrapPage(route)),
        );
      }

      res.end(promClient.register.metrics());

      next();
    }

    return middleware;
  })(),
);

const server = app.listen(process.env.PORT || 3000, () => {
  console.log('Server listening on port %d', server.address().port);
});
