/* eslint-disable no-console */
const express = require('express');
const bodyParser = require('body-parser');
const promClient = require('prom-client');
const yamlJs = require('yamljs');
const { get, merge, fromPairs, isArray } = require('lodash');
const extractMetrics = require('..');

class Store {
  constructor() {
    this.data = {};
  }

  set(key, value) {
    this.data[key] = value;
  }

  get(key) {
    return this.data[key];
  }
}

function getEnv(key, fallback) {
  return get(process.env, key, fallback);
}

let config = {
  path: getEnv('METRICS_PATH', '/metrics'),
  metricName: getEnv('METRIC_NAME', 'puppeteer_assets'),
  enableOnDemandQuery: getEnv('METRIC_ON_DEMAND_QUERY', false),
  defaults: {
    metrics: {
      file: getEnv('METRIC_FILE_ENABLED', true),
      count: getEnv('METRIC_COUNT_ENABLED', true),
      size: getEnv('METRIC_SIZE_ENABLED', true),
      gzip: getEnv('METRIC_GZIP_ENABLED', true),
      countByMimeType: getEnv('METRIC_COUNT_BY_MIME_TYPE_ENABLED', true),
      sizeByMimeType: getEnv('METRIC_SIZE_BY_MIME_TYPE_ENABLED', true),
      gzipSizeByMimeType: getEnv('METRIC_GZIP_SIZE_BY_MIME_TYPE_ENABLED', true),
    },
    mimeTypePatterns: [getEnv('METRIC_MIME_TYPE_PATTERN', 'javascript')],
    internalPatterns: [getEnv('METRIC_INTERNAL_PATTERN', '^(/|./)')],
  },
  configurations: [],
};

try {
  config = merge(
    {},
    config,
    yamlJs.load(process.env.CONFIG_PATH || `${__dirname}/../config.yml`) || {}
  );
  console.info('> Using configuration file...');
} catch (e) {
  config.enableOnDemandQuery = true;
  console.info('> No configuration file. On demand query enabled.');
}

function getConf(key, conf = {}) {
  return get(conf, key, get(config.defaults, key));
}

function parseLabelsParam(arr) {
  return isArray(arr) && arr.length
    ? fromPairs(arr.map(l => l.split(':'))) || {}
    : {};
}

function iterateToGauge(metrics, field, gauge, route) {
  const { url: page, labels } = route;
  Object.entries(metrics[field]).forEach(([type, data]) => {
    if (type === 'total') return;
    Object.entries(data).forEach(([mimeType, value]) => {
      if (mimeType === 'total') {
        gauge.set({ page, type, ...labels }, value);
      }
    });
  });
}

function iterateToGaugeByMimeType(metrics, field, gauge, route) {
  const { url: page, labels } = route;
  Object.entries(metrics[field]).forEach(([type, data]) => {
    if (type === 'total') return;
    Object.entries(data).forEach(([mimeType, value]) => {
      if (mimeType === 'total') return;
      gauge.set(
        {
          page,
          type,
          mime_type: mimeType,
          ...labels,
        },
        value
      );
    });
  });
}

const app = express();
app.use(bodyParser.json());

const store = new Store();

app.get(
  config.path,
  (() => {
    const { metricName } = config;

    const gaugeFile = getConf('metrics.file')
      ? new promClient.Gauge({
          name: `${metricName}_file`,
          help: 'historically found assets on scrap',
          labelNames: ['page', 'type', 'mime_type', 'url'],
        })
      : null;

    const gaugeCount = getConf('metrics.count')
      ? new promClient.Gauge({
          name: `${metricName}_count`,
          help: 'count of internal assets',
          labelNames: ['page', 'type'],
        })
      : null;

    const gaugeSize = getConf('metrics.size')
      ? new promClient.Gauge({
          name: `${metricName}_size`,
          help: 'real size of assets',
          labelNames: ['page', 'type'],
        })
      : null;

    const gaugeGzip = getConf('metrics.gzip')
      ? new promClient.Gauge({
          name: `${metricName}_gzip`,
          help: 'encoded (gzip) size of assets',
          labelNames: ['page', 'type'],
        })
      : null;

    const gaugeCountByMimeType = getConf('metrics.countByMimeType')
      ? new promClient.Gauge({
          name: `${metricName}_count_by_mime_type`,
          help: 'count of internal assets by mime type',
          labelNames: ['page', 'type', 'mime_type'],
        })
      : null;

    const gaugeSizeByMimeType = getConf('metrics.sizeByMimeType')
      ? new promClient.Gauge({
          name: `${metricName}_size_by_mime_type`,
          help: 'real size of assets by mime type',
          labelNames: ['page', 'type', 'mime_type'],
        })
      : null;

    const gaugeGzipByMimeType = getConf('metrics.gzipByMimeType')
      ? new promClient.Gauge({
          name: `${metricName}_gzip_by_mime_type`,
          help: 'encoded (gzip) size of assets by mime type',
          labelNames: ['page', 'type', 'mime_type'],
        })
      : null;

    const registerMetrics = async (metrics, route) => {
      promClient.register.setDefaultLabels(getConf('labels') || {});

      if (getConf('metrics.file')) {
        // file metrics
        const ts = Date.now();
        Object.entries(metrics.assets).forEach(([url, data]) => {
          const { type, mimeType } = data;
          gaugeFile.set(
            {
              page: route.url,
              url,
              type,
              mime_type: mimeType,
            },
            ts
          );
        });
      }

      // default
      if (getConf('metrics.count')) {
        iterateToGauge(metrics, 'count', gaugeCount, route);
      }
      if (getConf('metrics.size')) {
        iterateToGauge(metrics, 'size', gaugeSize, route);
      }
      if (getConf('metrics.gzip')) {
        iterateToGauge(metrics, 'gzip', gaugeGzip, route);
      }

      // by mime type
      if (getConf('metrics.countByMimeType')) {
        iterateToGaugeByMimeType(metrics, 'count', gaugeCountByMimeType, route);
      }
      if (getConf('metrics.sizeByMimeType')) {
        iterateToGaugeByMimeType(metrics, 'size', gaugeSizeByMimeType, route);
      }
      if (getConf('metrics.gzipByMimeType')) {
        iterateToGaugeByMimeType(metrics, 'gzip', gaugeGzipByMimeType, route);
      }
    };

    async function middleware(req, res, next) {
      if (req.query.url) {
        if (!config.enableOnDemandQuery) {
          res.status(400).send('On Demand is disabled');
          return;
        }
      } else if (config.enableOnDemandQuery && !config.configurations.length) {
        res.status(400).send('No URL query argument');
        return;
      } else if (!config.configurations.length) {
        res.status(400).send('No configurations');
        return;
      }

      promClient.register.resetMetrics();

      res.writeHead(200, { 'Content-Type': 'text/plain' });

      if (req.query.url) {
        const route = merge({}, config, {
          url: req.query.url,
          labels: parseLabelsParam(req.query.labels || []),
        });
        const metrics = await extractMetrics(route.url, route);
        registerMetrics(metrics, route);
      } else {
        await Promise.all(
          config.configurations.map(async route => {
            const metrics = store.get(route.url);
            console.log(!metrics && 'NO METRICS');
            if (metrics) registerMetrics(metrics, route);
          })
        );
      }

      res.end(promClient.register.metrics());

      next();
    }

    return middleware;
  })()
);

const server = app.listen(process.env.PORT || 3000, () => {
  console.log('Server listening on port %d', server.address().port);
});
