/* eslint-disable no-console */
const express = require('express');
const bodyParser = require('body-parser');
const promClient = require('prom-client');
const yamlJs = require('yamljs');
const { get, merge } = require('lodash');
const Store = require('./store');
const {
  getEnv,
  parseLabelsParam,
  parseMimeTypesParam,
} = require('./utils');
const getMetrics = require('..');

const DEFAULT_MIMETYPES = ['javascript'];

let Settings = {
  path: getEnv('METRICS_PATH', '/metrics'),
  metricName: getEnv('METRIC_NAME', 'puppeteer_assets'),
  enableOnDemandQuery: getEnv('METRIC_ON_DEMAND_QUERY', false),
  defaults: {
    metrics: {
      file: getEnv('METRIC_FILE_ENABLED', true),
      count: getEnv('METRIC_COUNT_ENABLED', true),
      size: getEnv('METRIC_SIZE_ENABLED', true),
      gzip: getEnv('METRIC_GZIP_ENABLED', true),
      countByMimeType: getEnv('METRIC_COUNT_BY_MIMETYPE_ENABLED', true),
      sizeByMimeType: getEnv('METRIC_SIZE_BY_MIMETYPE_ENABLED', true),
      gzipSizeByMimeType: getEnv('METRIC_GZIP_SIZE_BY_MIMETYPE_ENABLED', true),
    },
    mimeTypes: getEnv('METRIC_MIMETYPES') ? getEnv('METRIC_MIMETYPES').split(',') : DEFAULT_MIMETYPES,
  },
  configurations: [],
};

try {
  Settings = merge(
    {},
    Settings,
    yamlJs.load(process.env.CONFIG_PATH || `${__dirname}/../config.yml`) || {},
  );
  console.info('> Using configuration file...');
} catch (e) {
  Settings.enableOnDemandQuery = true;
  console.info('> No configuration file. On demand query enabled.');
}

function getConf(key, conf = {}) {
  return get(conf, key, get(Settings.defaults, key));
}

function iterateToGauge(metrics, field, gauge, conf) {
  const { url: page, labels } = conf;
  const mimeTypes = getConf('mimeTypes', conf);
  Object.entries(metrics[field]).forEach(([type, data]) => {
    if (type === 'total') return;
    mimeTypes.forEach((mimeType) => {
      gauge.set({ page, type, ...labels }, get(data[mimeType], 'total') || 0);
    });
  });
}

function iterateToGaugeByMimeType(metrics, field, gauge, conf) {
  const { url: page, labels } = conf;
  const mimeTypes = getConf('mimeTypes', conf);
  Object.entries(metrics[field]).forEach(([type, data]) => {
    if (type === 'total') return;
    mimeTypes.forEach((mimeType) => {
      gauge.set({
        page, type, mime_type: mimeType, ...labels,
      }, get(data[mimeType], 'total') || 0);
    });
  });
}

function stripInvalidLabels(labelKeys, conf) {
  const labels = labelKeys.reduce((acc, key) => {
    const value = getConf('labels', conf)[key];
    acc[key] = value || null;
    return acc;
  }, {});
  return { ...conf, labels };
}

function assertRouting(req) {
  if (req.query.url) {
    if (!Settings.enableOnDemandQuery) {
      throw new Error('On Demand is disabled');
    }
  } else if (Settings.enableOnDemandQuery && !Settings.configurations.length) {
    throw new Error('No URL query argument');
  } else if (!Settings.configurations.length) {
    throw new Error('No configurations');
  }
}

const app = express();
const store = new Store();

app.use(bodyParser.json());

function registerMetrics(metrics, conf, {
  gaugeFile,
  gaugeCount,
  gaugeSize,
  gaugeGzip,
  gaugeCountByMimeType,
  gaugeSizeByMimeType,
  gaugeGzipByMimeType,
}) {
  const labels = getConf('labels', conf);

  promClient.register.setDefaultLabels(labels);

  if (getConf('metrics.file', conf)) {
    // file metrics
    const ts = Date.now();
    Object.entries(metrics.assets).forEach(([url, data]) => {
      const { type, mimeType } = data;
      gaugeFile.set(
        {
          page: conf.url,
          url,
          type,
          mime_type: mimeType,
        },
        ts,
      );
    });
  }

  // default
  if (getConf('metrics.count', conf)) {
    iterateToGauge(metrics, 'count', gaugeCount, conf);
  }
  if (getConf('metrics.size', conf)) {
    iterateToGauge(metrics, 'size', gaugeSize, conf);
  }
  if (getConf('metrics.gzip', conf)) {
    iterateToGauge(metrics, 'gzip', gaugeGzip, conf);
  }

  // by mime type
  if (getConf('metrics.countByMimeType', conf)) {
    iterateToGaugeByMimeType(metrics, 'count', gaugeCountByMimeType, conf);
  }
  if (getConf('metrics.sizeByMimeType', conf)) {
    iterateToGaugeByMimeType(metrics, 'size', gaugeSizeByMimeType, conf);
  }
  if (getConf('metrics.gzipByMimeType', conf)) {
    iterateToGaugeByMimeType(metrics, 'gzip', gaugeGzipByMimeType, conf);
  }
}

function setupMetrics(configurations, overrideSettings = {}) {
  const config = { ...Settings, ...overrideSettings };
  const { metricName, labels: labelKeys } = config;

  promClient.register.clear();

  const gaugeUp = new promClient.Gauge({
    name: `${metricName}_up`,
    help: '1 = up, 0 = not up',
    labelNames: labelKeys,
  });

  const gaugeFile = getConf('metrics.file')
    ? new promClient.Gauge({
      name: `${metricName}_file`,
      help: 'historically found assets on scrap',
      labelNames: ['page', 'type', 'mime_type', 'url', ...labelKeys],
    })
    : null;

  const gaugeCount = getConf('metrics.count')
    ? new promClient.Gauge({
      name: `${metricName}_count`,
      help: 'count of internal assets',
      labelNames: ['page', 'type', ...labelKeys],
    })
    : null;

  const gaugeSize = getConf('metrics.size')
    ? new promClient.Gauge({
      name: `${metricName}_size`,
      help: 'real size of assets',
      labelNames: ['page', 'type', ...labelKeys],
    })
    : null;

  const gaugeGzip = getConf('metrics.gzip')
    ? new promClient.Gauge({
      name: `${metricName}_gzip`,
      help: 'encoded (gzip) size of assets',
      labelNames: ['page', 'type', ...labelKeys],
    })
    : null;

  const gaugeCountByMimeType = getConf('metrics.countByMimeType')
    ? new promClient.Gauge({
      name: `${metricName}_count_by_mime_type`,
      help: 'count of internal assets by mime type',
      labelNames: ['page', 'type', 'mime_type', ...labelKeys],
    })
    : null;

  const gaugeSizeByMimeType = getConf('metrics.sizeByMimeType')
    ? new promClient.Gauge({
      name: `${metricName}_size_by_mime_type`,
      help: 'real size of assets by mime type',
      labelNames: ['page', 'type', 'mime_type', ...labelKeys],
    })
    : null;

  const gaugeGzipByMimeType = getConf('metrics.gzipByMimeType')
    ? new promClient.Gauge({
      name: `${metricName}_gzip_by_mime_type`,
      help: 'encoded (gzip) size of assets by mime type',
      labelNames: ['page', 'type', 'mime_type', ...labelKeys],
    })
    : null;

  configurations.forEach((rawConf) => {
    const conf = stripInvalidLabels(labelKeys, rawConf);
    const metrics = store.get(conf.url);

    gaugeUp.set(getConf('labels', conf), metrics ? 1 : 0);

    if (metrics) {
      registerMetrics(metrics, conf, {
        gaugeFile,
        gaugeCount,
        gaugeSize,
        gaugeGzip,
        gaugeCountByMimeType,
        gaugeSizeByMimeType,
        gaugeGzipByMimeType,
      });
      return true;
    }
    return false;
  });
}

function configureTimer() {
  if (!Settings.configurations.length) return;

  setInterval(async () => {
    console.info('> timer: scraping pages...');
    store.busy = true;
    await Promise.all(
      Settings.configurations.map(async (conf) => {
        const metrics = await getMetrics(conf.url, conf);
        if (metrics) store.set(conf.url, metrics);
      }),
    );
    store.busy = false;
    console.info('> timer: done...');
  }, 30000);
}

app.get(Settings.path, async (req, res, next) => {
  try {
    assertRouting(req);
  } catch (err) {
    res.status(400).send(err.message);
    return;
  }

  // on demand query
  try {
    if (req.query.url) {
      if (store.busy) throw new Error('Ops.. Busy service!');

      const conf = merge({}, Settings.defaults);
      conf.url = req.query.url;
      conf.labels = parseLabelsParam(req.query.labels || []);
      conf.mimeTypes = parseMimeTypesParam(req.query.mimeTypes || DEFAULT_MIMETYPES);

      const metrics = await getMetrics(conf.url, conf);
      store.set(conf.url, metrics);

      const overrideSettings = { labels: Object.keys(conf.labels) };
      setupMetrics([conf], overrideSettings);
    } else {
      setupMetrics(Settings.configurations);
    }
  } catch (err) {
    res.status(500).send(`ERROR: ${err.message}`);
    throw err;
  }

  res.writeHead(200, { 'Content-Type': 'text/plain' });
  res.end(promClient.register.metrics());
  next();
});

const server = app.listen(process.env.PORT || 3000, () => {
  configureTimer();
  console.log('Server listening on port %d', server.address().port);
});
